@@ -1,21 +1,26 @@
 # -*- coding: utf-8 -*-
 import asyncio
+import errno
 import datetime
 from contextlib import contextmanager
 import json
 from concurrent.futures import ProcessPoolExecutor
 import io
 import os
+import random
 import requests
 import time
 import sys
 import pytest
+import moto
 from itertools import chain
 import fsspec.core
+import s3fs.core
 from s3fs.core import S3FileSystem
 from s3fs.utils import ignoring, SSEParams
 from botocore.exceptions import NoCredentialsError
 from fsspec.asyn import sync
+from packaging import version
 
 test_bucket_name = "test"
 secure_bucket_name = "test-secure"
@@ -55,7 +60,6 @@ a = test_bucket_name + "/tmp/test/a"
 b = test_bucket_name + "/tmp/test/b"
 c = test_bucket_name + "/tmp/test/c"
 d = test_bucket_name + "/tmp/test/d"
-py35 = sys.version_info[:2] == (3, 5)
 port = 5555
 endpoint_uri = "http://127.0.0.1:%s/" % port
 
@@ -83,13 +87,17 @@ def s3_base():
     proc.wait()
 
 
-@pytest.fixture()
-def s3(s3_base):
+def get_boto3_client():
     from botocore.session import Session
 
     # NB: we use the sync botocore client for setup
     session = Session()
-    client = session.create_client("s3", endpoint_url=endpoint_uri)
+    return session.create_client("s3", endpoint_url=endpoint_uri)
+
+
+@pytest.fixture()
+def s3(s3_base):
+    client = get_boto3_client()
     client.create_bucket(Bucket=test_bucket_name, ACL="public-read")
 
     client.create_bucket(Bucket=versioned_bucket_name, ACL="public-read")
@@ -209,7 +217,7 @@ def test_config_kwargs_class_attributes_override():
 
 def test_idempotent_connect(s3):
     first = s3.s3
-    assert s3.connect() is not first
+    assert s3.connect(refresh=True) is not first
 
 
 def test_multiple_objects(s3):
@@ -414,7 +422,7 @@ def test_not_delegate():
 
 
 def test_ls(s3):
-    assert set(s3.ls("")) == {
+    assert set(s3.ls("", detail=False)) == {
         test_bucket_name,
         secure_bucket_name,
         versioned_bucket_name,
@@ -422,7 +430,7 @@ def test_ls(s3):
     with pytest.raises(FileNotFoundError):
         s3.ls("nonexistent")
     fn = test_bucket_name + "/test/accounts.1.json"
-    assert fn in s3.ls(test_bucket_name + "/test")
+    assert fn in s3.ls(test_bucket_name + "/test", detail=False)
 
 
 def test_pickle(s3):
@@ -776,10 +784,11 @@ def test_copy_managed(s3):
         sync(s3.loop, s3._copy_managed, fn, fn + "3", size=len(data), block=6 * 2 ** 30)
 
 
-def test_move(s3):
+@pytest.mark.parametrize("recursive", [True, False])
+def test_move(s3, recursive):
     fn = test_bucket_name + "/test/accounts.1.json"
     data = s3.cat(fn)
-    s3.mv(fn, fn + "2")
+    s3.mv(fn, fn + "2", recursive=recursive)
     assert s3.cat(fn + "2") == data
     assert not s3.exists(fn)
 
@@ -855,6 +864,26 @@ def test_errors(s3):
         s3.find("s3://")
 
 
+def test_errors_cause_preservings(monkeypatch, s3):
+    # We translate the error, and preserve the original one
+    with pytest.raises(FileNotFoundError) as exc:
+        s3.rm("unknown")
+
+    assert type(exc.value.__cause__).__name__ == "NoSuchBucket"
+
+    async def head_object(*args, **kwargs):
+        raise NoCredentialsError
+
+    monkeypatch.setattr(type(s3.s3), "head_object", head_object)
+
+    # Since the error is not translate, the __cause__ would
+    # be None
+    with pytest.raises(NoCredentialsError) as exc:
+        s3.info("test/a.txt")
+
+    assert exc.value.__cause__ is None
+
+
 def test_read_small(s3):
     fn = test_bucket_name + "/2014-01-01.csv"
     with s3.open(fn, "rb", block_size=10) as f:
@@ -1188,7 +1217,19 @@ def _get_s3_id(s3):
 
 
 @pytest.mark.skipif(sys.version_info[:2] < (3, 7), reason="ctx method only >py37")
-@pytest.mark.parametrize("method", ["spawn", "forkserver"])
+@pytest.mark.parametrize(
+    "method",
+    [
+        "spawn",
+        pytest.param(
+            "forkserver",
+            marks=pytest.mark.skipif(
+                sys.platform.startswith("win"),
+                reason="'forserver' not available on windows",
+            ),
+        ),
+    ],
+)
 def test_no_connection_sharing_among_processes(s3, method):
     import multiprocessing as mp
 
@@ -1279,7 +1320,6 @@ def test_tags(s3):
     assert s3.get_tags(fname) == tagset
 
 
-@pytest.mark.skipif(py35, reason="no versions on old moto for py36")
 def test_versions(s3):
     versioned_file = versioned_bucket_name + "/versioned_file"
     s3 = S3FileSystem(
@@ -1287,23 +1327,29 @@ def test_versions(s3):
     )
     with s3.open(versioned_file, "wb") as fo:
         fo.write(b"1")
+    first_version = fo.version_id
+
     with s3.open(versioned_file, "wb") as fo:
         fo.write(b"2")
+    second_version = fo.version_id
+
     assert s3.isfile(versioned_file)
     versions = s3.object_version_info(versioned_file)
-    version_ids = [version["VersionId"] for version in versions]
-    assert len(version_ids) == 2
+    assert len(versions) == 2
+    assert {version["VersionId"] for version in versions} == {
+        first_version,
+        second_version,
+    }
 
     with s3.open(versioned_file) as fo:
-        assert fo.version_id == version_ids[1]
+        assert fo.version_id == second_version
         assert fo.read() == b"2"
 
-    with s3.open(versioned_file, version_id=version_ids[0]) as fo:
-        assert fo.version_id == version_ids[0]
+    with s3.open(versioned_file, version_id=first_version) as fo:
+        assert fo.version_id == first_version
         assert fo.read() == b"1"
 
 
-@pytest.mark.skipif(py35, reason="no versions on old moto for py36")
 def test_list_versions_many(s3):
     # moto doesn't actually behave in the same way that s3 does here so this doesn't test
     # anything really in moto 1.2
@@ -1345,7 +1391,6 @@ def test_fsspec_versions_multiple(s3):
             assert contents == version_lookup[fo.version_id]
 
 
-@pytest.mark.skipif(py35, reason="no versions on old moto for py36")
 def test_versioned_file_fullpath(s3):
     versioned_file = versioned_bucket_name + "/versioned_file_fullpath"
     s3 = S3FileSystem(
@@ -1404,10 +1449,10 @@ def test_text_io__basic(s3):
     """Text mode is now allowed."""
     s3.mkdir("bucket")
 
-    with s3.open("bucket/file.txt", "w") as fd:
+    with s3.open("bucket/file.txt", "w", encoding="utf-8") as fd:
         fd.write("\u00af\\_(\u30c4)_/\u00af")
 
-    with s3.open("bucket/file.txt", "r") as fd:
+    with s3.open("bucket/file.txt", "r", encoding="utf-8") as fd:
         assert fd.read() == "\u00af\\_(\u30c4)_/\u00af"
 
 
@@ -1544,32 +1589,33 @@ def test_touch(s3):
     assert s3.size(fn) == 4
 
 
-@pytest.mark.skipif(py35, reason="no versions on old moto for py36")
 def test_touch_versions(s3):
     versioned_file = versioned_bucket_name + "/versioned_file"
     s3 = S3FileSystem(
         anon=False, version_aware=True, client_kwargs={"endpoint_url": endpoint_uri}
     )
-    returned_versions = []
+
     with s3.open(versioned_file, "wb") as fo:
         fo.write(b"1")
-    returned_versions.append(fo.version_id)
+    first_version = fo.version_id
     with s3.open(versioned_file, "wb") as fo:
         fo.write(b"")
-    returned_versions.append(fo.version_id)
+    second_version = fo.version_id
+
     assert s3.isfile(versioned_file)
     versions = s3.object_version_info(versioned_file)
-    version_ids = [version["VersionId"] for version in versions]
-    assert len(version_ids) == 2
+    assert len(versions) == 2
+    assert {version["VersionId"] for version in versions} == {
+        first_version,
+        second_version,
+    }
 
     with s3.open(versioned_file) as fo:
-        assert fo.version_id == version_ids[1]
-        assert fo.version_id == returned_versions[1]
+        assert fo.version_id == second_version
         assert fo.read() == b""
 
-    with s3.open(versioned_file, version_id=version_ids[0]) as fo:
-        assert fo.version_id == version_ids[0]
-        assert fo.version_id == returned_versions[0]
+    with s3.open(versioned_file, version_id=first_version) as fo:
+        assert fo.version_id == first_version
         assert fo.read() == b"1"
 
 
@@ -1673,8 +1719,9 @@ def test_credentials():
     )
     assert s3.s3._request_signer._credentials.access_key == "foobar"
     assert s3.s3._request_signer._credentials.secret_key == "foobar"
-    with pytest.raises(TypeError) as excinfo:
-        s3 = S3FileSystem(
+    with pytest.raises((TypeError, KeyError)):
+        # should be TypeError: arg passed twice; but in moto can be KeyError
+        S3FileSystem(
             key="foo",
             secret="foo",
             client_kwargs={
@@ -1682,8 +1729,7 @@ def test_credentials():
                 "aws_secret_access_key": "bar",
                 "endpoint_url": endpoint_uri,
             },
-        )
-        assert "multiple values for keyword argument" in str(excinfo.value)
+        ).s3
 
 
 def test_modified(s3):
@@ -1717,18 +1763,41 @@ def test_async_s3(s3):
         fn = test_bucket_name + "/nested/file1"
         data = b"hello\n"
 
-        # Fails because client creation has not yet been awaited
-        with pytest.raises(RuntimeError):
-            await s3._cat_file(fn)
+        # Is good with or without connect()
+        await s3._cat_file(fn)
 
-        s3.connect()  # creates client
+        session = await s3.set_session()  # creates client
 
         assert await s3._cat_file(fn) == data
 
-        with s3.open(fn, "rb") as f:
-            assert f.read() == data
+        assert await s3._cat_file(fn, start=0, end=3) == data[:3]
 
-        await s3._s3.close()
+        # TODO: file IO is *not* async
+        # with s3.open(fn, "rb") as f:
+        #     assert f.read() == data
+
+        await session.close()
+
+    asyncio.run(_())
+
+
+@pytest.mark.skipif(sys.version_info < (3, 7), reason="no asyncio.run in py36")
+def test_async_s3_old(s3):
+    async def _():
+        s3 = S3FileSystem(
+            anon=False,
+            asynchronous=True,
+            loop=asyncio.get_running_loop(),
+            client_kwargs={"region_name": "eu-central-1", "endpoint_url": endpoint_uri},
+        )
+
+        fn = test_bucket_name + "/nested/file1"
+        data = b"hello\n"
+
+        # Check old API
+        session = await s3._connect()
+        assert await s3._cat_file(fn, start=0, end=3) == data[:3]
+        await session.close()
 
     asyncio.run(_())
 
@@ -1825,5 +1894,129 @@ def test_version_sizes(s3):
     for version in versions:
         version_id = version["VersionId"]
         with s3.open(path, version_id=version_id) as f:
-            with gzip.open(f) as zfp:
+            with gzip.GzipFile(fileobj=f) as zfp:
                 zfp.read()
+
+
+def test_find_no_side_effect(s3):
+    infos1 = s3.find(test_bucket_name, maxdepth=1, withdirs=True, detail=True)
+    s3.find(test_bucket_name, maxdepth=None, withdirs=True, detail=True)
+    infos3 = s3.find(test_bucket_name, maxdepth=1, withdirs=True, detail=True)
+    assert infos1.keys() == infos3.keys()
+
+
+def test_get_file_info_with_selector(s3):
+    fs = s3
+    base_dir = "selector-dir/"
+    file_a = "selector-dir/test_file_a"
+    file_b = "selector-dir/test_file_b"
+    dir_a = "selector-dir/test_dir_a"
+    file_c = "selector-dir/test_dir_a/test_file_c"
+
+    try:
+        fs.mkdir(base_dir)
+        with fs.open(file_a, mode="wb"):
+            pass
+        with fs.open(file_b, mode="wb"):
+            pass
+        fs.mkdir(dir_a)
+        with fs.open(file_c, mode="wb"):
+            pass
+
+        infos = fs.find(base_dir, maxdepth=None, withdirs=True, detail=True)
+        assert len(infos) == 4
+
+        for info in infos.values():
+            if info["name"].endswith(file_a):
+                assert info["type"] == "file"
+            elif info["name"].endswith(file_b):
+                assert info["type"] == "file"
+            elif info["name"].endswith(file_c):
+                assert info["type"] == "file"
+            elif info["name"].rstrip("/").endswith(dir_a):
+                assert info["type"] == "directory"
+            else:
+                raise ValueError("unexpected path {}".format(info["name"]))
+    finally:
+        fs.rm(base_dir, recursive=True)
+
+
+@pytest.mark.xfail(
+    condition=version.parse(moto.__version__) <= version.parse("1.3.16"),
+    reason="Moto 1.3.16 is not supporting pre-conditions.",
+)
+def test_raise_exception_when_file_has_changed_during_reading(s3):
+    test_file_name = "file1"
+    test_file = "s3://" + test_bucket_name + "/" + test_file_name
+    content1 = b"123"
+    content2 = b"ABCDEFG"
+
+    boto3_client = get_boto3_client()
+
+    def create_file(content: bytes):
+        boto3_client.put_object(
+            Bucket=test_bucket_name, Key=test_file_name, Body=content
+        )
+
+    create_file(b"123")
+
+    with s3.open(test_file, "rb") as f:
+        content = f.read()
+        assert content == content1
+
+    with s3.open(test_file, "rb") as f:
+        create_file(content2)
+        with expect_errno(errno.EBUSY):
+            f.read()
+
+
+def test_s3fs_etag_preserving_multipart_copy(monkeypatch, s3):
+    # Set this to a lower value so that we can actually
+    # test this without creating giant objects in memory
+    monkeypatch.setattr(s3fs.core, "MANAGED_COPY_THRESHOLD", 5 * 2 ** 20)
+
+    test_file1 = test_bucket_name + "/test/multipart-upload.txt"
+    test_file2 = test_bucket_name + "/test/multipart-upload-copy.txt"
+
+    with s3.open(test_file1, "wb", block_size=5 * 2 ** 21) as stream:
+        for _ in range(5):
+            stream.write(b"b" * (stream.blocksize + random.randrange(200)))
+
+    file_1 = s3.info(test_file1)
+
+    s3.copy(test_file1, test_file2)
+    file_2 = s3.info(test_file2)
+    s3.rm(test_file2)
+
+    # normal copy() uses a block size of 5GB
+    assert file_1["ETag"] != file_2["ETag"]
+
+    s3.copy(test_file1, test_file2, preserve_etag=True)
+    file_2 = s3.info(test_file2)
+    s3.rm(test_file2)
+
+    # etag preserving copy() determines each part size for the destination
+    # by checking out the matching part's size on the source
+    assert file_1["ETag"] == file_2["ETag"]
+
+    s3.rm(test_file1)
+
+
+@pytest.mark.skipif(sys.version_info < (3, 7), reason="no asyncio.run in py36")
+def test_sync_from_wihin_async(s3):
+    # if treating as sync but within an even loop, e.g., calling from jupyter;
+    # IO happens on dedicated thread.
+    async def f():
+        S3FileSystem.clear_instance_cache()
+        s3 = S3FileSystem(anon=False, client_kwargs={"endpoint_url": endpoint_uri})
+        assert s3.ls(test_bucket_name)
+
+    asyncio.run(f())
+
+
+def test_token_paths(s3):
+    fs, tok, files = fsspec.get_fs_token_paths(
+        "s3://" + test_bucket_name + "/*.csv",
+        storage_options={"client_kwargs": {"endpoint_url": endpoint_uri}},
+    )
+    assert files
